@@ -1,76 +1,94 @@
 package Spreadsheet::HTML;
 use strict;
 use warnings FATAL => 'all';
-our $VERSION = '0.07';
+our $VERSION = '0.08';
 
-use Math::Matrix;
+use Clone;
 use HTML::Element;
+use Math::Matrix;
 
-sub new {
-    my $class = shift;
-    my %attrs = ref($_[0]) eq 'HASH' ? %{+shift} : @_;
-    return bless { %attrs }, $class;
-}
+use Spreadsheet::HTML::CSV;
+use Spreadsheet::HTML::HTML;
+use Spreadsheet::HTML::JSON;
+use Spreadsheet::HTML::YAML;
 
-sub generate    { _wrapper( sub { @_ }, @_ ) }
-sub transpose   { _wrapper( sub { @{ Math::Matrix::transpose( [@_] ) } }, @_ ) }
-sub reverse     { _wrapper( sub { reverse @_ }, @_ ) }
 sub portrait    { generate( @_ ) }
+sub generate    { _make_table( process( @_ ) ) }
+
 sub landscape   { transpose( @_ ) }
-
-sub _wrapper {
-    my $sub   = shift;
-    my $self  = shift if UNIVERSAL::isa( $_[0], __PACKAGE__ );
-    my $attrs = ref($_[0]) eq 'HASH' ? shift : $self ? {%$self} : {};
-
-    if (@_ > 1 && defined($_[0]) && !ref($_[0]) ) {
-        my %args = @_;
-        @_ = delete $args{data};
-        $attrs = {%args};    
-    }
-
-    my @data  = $self ? $self->process_data( @_ ) : process_data( @_ );
-    return _make_table( $attrs, $sub->( @data ) );
+sub transpose   {
+    my %args = process( @_ );
+    $args{data} = [@{ Math::Matrix::transpose( $args{data} ) }];
+    return _make_table( %args );
 }
 
-sub process_data {
-    my ($self, @data);
+sub flip   {
+    my %args = process( @_ );
+    $args{data} = [ CORE::reverse @{ $args{data} } ];
+    return _make_table( %args );
+}
 
-	if (UNIVERSAL::isa( $_[0], __PACKAGE__ )) {
-        $self = shift;
-        return @{ $self->{data} } if exists $self->{__processed_data__};
-        $self->{table_attrs} = shift if ref($_[0]) eq 'HASH';
-        @_ = $self->{data};
+sub mirror   {
+    my %args = process( @_ );
+    $args{data} = [ map [ CORE::reverse @$_ ], @{ $args{data} } ];
+    return _make_table( %args );
+}
+
+sub reverse   {
+    my %args = process( @_ );
+    $args{data} = [ map [ CORE::reverse @$_ ], CORE::reverse @{ $args{data} } ];
+    return _make_table( %args );
+}
+
+sub process {
+    my ($self,$data,$args) = _args( @_ );
+
+    $data = Clone::clone( $self->{data} ) if $self->{is_cached};
+
+    my $max_cols = scalar @{ $data->[0] };
+
+    for my $row (@$data) {
+        push @$row, undef for 1 .. ($max_cols - scalar @$row);
+        #truncate rows that are too long
+        $row = [ @$row[0 .. $max_cols - 1] ] if scalar( @$row ) > $max_cols;
+
+        # white space transliteration
+        for my $i (0 .. $#$row) {
+            #TODO: this needs to be configurable
+            do{ no warnings; $row->[$i] =~ s/^\s*$/&nbsp;/g };
+            $row->[$i] =~ s/\n/<br \/>/g;
+        }
     }
 
-    @data = @_ > 1 ? @_ : @{ ref($_[0]) ? $_[0] : [ $_[0] ] };
-    @data = [ @data ] unless ref( $data[0] ) eq 'ARRAY';
-    @data = [ undef ] unless @{ $data[0] };
+    unless ( $args->{headless} or $args->{matrix} or ref($data->[0][0]) ) {
+        #TODO: make objects here, assign class attrs here
+        $data->[0] = [ map [$_], @{ $data->[0] } ];
+    }
 
-    $self->{data} = _process( @data );
-    $self->{data} = _mark_headings( $self->{data} );
-    $self->{__processed_data__} = 1;
-    return @{ $self->{data} };
+    if ($args->{headless}) {
+        shift @$data;
+    }
+
+    return wantarray ? ( data => $data, %$args ) : $data;
 }
 
 sub _make_table {
-    my $attrs = ref($_[0]) eq 'HASH' ? shift : {};
-    $attrs->{$_} ||= {} for qw( table tr th td );
+    my %args = @_;
+    $args{$_} ||= {} for qw( table tr th td );
 
-    my $indent  = delete $attrs->{indent};
-    my $no_th   = delete $attrs->{matrix};
-    my $encodes = '';
-    $encodes = delete $attrs->{encodes} if exists $attrs->{encodes};
+    my $indent  = $args{indent};
+    my $no_th   = $args{matrix};
+    my $encodes = exists $args{encodes} ? $args{encodes} : '';
 
     my $table = HTML::Element->new_from_lol(
-        [table => $attrs->{table},
-            map [tr => $attrs->{tr},
-                map ref($_) 
+        [table => $args{table},
+            map [tr => $args{tr},
+                map ref($_)
                     ? $no_th
-                        ? [ td => $attrs->{td}, @$_ ]
-                        : [ th => $attrs->{th}, @$_ ]
-                    : [ td => $attrs->{td}, $_ ], @$_
-            ], @_
+                        ? [ td => $args{td}, @$_ ]
+                        : [ th => $args{th}, @$_ ]
+                    : [ td => $args{td}, $_ ], @$_
+            ], @{ $args{data} }
         ],
     );
 
@@ -78,32 +96,60 @@ sub _make_table {
     return $html;
 }
 
-# just a way to seperate th cells from td cells now (for later)
-sub _mark_headings {
-    my $data = shift;
-    $data->[0] = [ map [$_], @{ $data->[0] } ];
-    return $data;
+sub new {
+    my $class = shift;
+    my %attrs = ref($_[0]) eq 'HASH' ? %{+shift} : @_;
+    return bless { %attrs }, $class;
 }
 
+sub _args {
+    my ($self,$data,$args);
+    $self = shift if UNIVERSAL::isa( $_[0], __PACKAGE__ );
 
-sub _process {
-    my @data = @_;
-
-    # padding is determined by first row (the headings)
-    my $max_cols = scalar @{ $data[0] };
-    for my $row (@data[1 .. $#data]) {
-        push @$row, undef for 1 .. ($max_cols - scalar @$row);
+    if (@_ > 1 && defined($_[0]) && !ref($_[0]) ) {
+        my %args = @_;
+        if (my $arg = delete $args{data}) {
+            $data = $arg;
+        }
+        $args = {%args};
+    } elsif (@_ > 1 && ref($_[0]) eq 'ARRAY') {
+        $data = [ @_ ];
+    } elsif (@_ == 1) {
+        $data = $_[0];
     }
 
-    return [
-        map { [
-            map {
-                do { no warnings; s/^\s*$/&nbsp;/g };
-                s/\n/<br \/>/g;
-                $_
-            } @$_
-        ] } @data
-    ];
+    if ($self) {
+        $args = { %$self, %{ $args || {} } };
+        delete $args->{data};
+    }
+
+    unless (exists $args->{file}) {
+        if ($self and !$data) {
+            $data = $self->{data};
+        }
+        $data = [ $data ] unless ref($data);
+        $data = [ $data ] unless ref($data->[0]);
+        $data = [ [undef] ] if !scalar @{ $data->[0] };
+    }
+
+    if (my $file = $args->{file}) {
+        if ($file =~ /\.csv$/) {
+            $data = Spreadsheet::HTML::CSV::load( $file );
+        } elsif ($file =~ /\.html?$/) {
+            $data = Spreadsheet::HTML::HTML::load( $file );
+        } elsif ($file =~ /\.jso?n$/) {
+            $data = Spreadsheet::HTML::JSON::load( $file );
+        } elsif ($file =~ /\.ya?ml$/) {
+            $data = Spreadsheet::HTML::YAML::load( $file );
+        }
+    }
+
+    if ($self and !$self->{is_cached} and delete $args->{cache}) {
+        $self->{data} = $data;
+        $self->{is_cached} = 1;
+    }
+
+    return ( $self, Clone::clone($data), $args );
 }
 
 1;
@@ -111,7 +157,7 @@ sub _process {
 __END__
 =head1 NAME
 
-Spreadsheet::HTML - Tabular data to HTML tables.
+Spreadsheet::HTML - Render HTML tables with ease.
 
 =head1 THIS IS AN ALPHA RELEASE.
 
@@ -121,7 +167,7 @@ todo:
 
 =over 4
 
-=item * emit col, colgroup, thead, tbody and caption tags
+=item * emit col, colgroup, thead, tbody and caption tags ... maybe ...
 
 =item * map client functions to cells
 
@@ -134,20 +180,6 @@ todo:
 You are encouraged to try my older L<DBIx::XHTML_Table> during
 the development of this module.
 
-=head1 REQUIRES
-
-=over 4
-
-=item HTML::Entities
-
-Used to encode values with HTML entities.
-
-=item Math::Matrix
-
-Used to transpose data from portrait to landscape.
-
-=back
-
 =head1 SYNOPSIS
 
     use Spreadsheet::HTML;
@@ -159,14 +191,18 @@ Used to transpose data from portrait to landscape.
         [qw(col1 col2 col3)],
     ];
 
-    my $table = Spreadsheet::HTML->new( data => $data );
+    my $table = Spreadsheet::HTML->new( data => $data, cache => 1 );
     print $table->generate;
     print $table->transpose;
+    print $table->flip;
+    print $table->mirror;
     print $table->reverse;
 
     # non OO
     print Spreadsheet::HTML::generate( $data );
     print Spreadsheet::HTML::transpose( $data );
+    print Spreadsheet::HTML::flip( $data );
+    print Spreadsheet::HTML::mirror( $data );
     print Spreadsheet::HTML::reverse( $data );
 
 =head1 METHODS
@@ -179,6 +215,7 @@ Used to transpose data from portrait to landscape.
   my $table = Spreadsheet::HTML->new( @data );
   my $table = Spreadsheet::HTML->new( $data );
   my $table = Spreadsheet::HTML->new( data => $data );
+  my $table = Spreadsheet::HTML->new( { data => $data } );
 
 Constructs object. Currently accepts one parameter: data.
 Data should be a two dimensional array and you should 
@@ -186,9 +223,11 @@ expect the first row to be treated as the header (which
 means each cell will be wrapped with <th> tags instead 
 of <td> tags).
 
-=item generate()
+=item process( key => 'value' )
 
-=item portrait()
+Data structure that can be used by the following:
+
+=item generate( key => 'value' )
 
   my $html = $table->generate;
   my $html = $table->generate( indent => '    ' );
@@ -216,35 +255,47 @@ all data will:
 
 =back
 
-=item transpose()
+These features are currently hard coded in (sorry). Plans
+to make these transliterations configurable by the client
+are planned. Plans planning plans.
 
-=item landscape()
+=item portrait( key => 'value' )
 
-  (same usage as generate)    
+Alias for generate()
 
-Uses Math::Matrix to rotate the data 90 degrees and
-then returns a string containing the rendered HTML table.
+=item transpose( key => 'value' )
 
-=item reverse()
+Uses Math::Matrix to rotate the headings and data
+90 degrees counter-clockwise.
 
-  (same usage as generate)    
+=item landscape( key => 'value' )
 
-Uses Math::Matrix to rotate the data 180 degrees and
-then returns a string containing the rendered HTML table.
+Alias for transpose()
 
-=item process_data()
+=item flip( key => 'value' )
 
-Returns the munged data before it is used to generate
-a rendered HTML table. This is called on your behalf, but
-is available should you want it.
+Flips the headings and data upside down.
+
+=item mirror( key => 'value' )
+
+Columns are rendered right to left.
+
+=item reverse( key => 'value' )
+
+Combines flip and mirror: flips the headings and
+data upside down and render columns right to left.
 
 =back
 
 =head1 ATTRIBUTES
 
-All methods/procedures can accept named arguments.
-If named arguments are detected, the data has to be
-an array ref assigned to the key 'data'.
+All methods/procedures accept named arguments.
+If named arguments are detected: the data has to be
+an array ref assigned to the key 'data'. If no
+named args are detected then the parameter list is
+treated as the data itself, either an array containing
+array references or an array reference containing
+array references.
 
 =over 4
 
@@ -252,25 +303,39 @@ an array ref assigned to the key 'data'.
 
 The data to be rendered into table cells.
 
-=item * indent => 0 or 1
+=item * file => $str
+
+The name of the data file to read. Supported formats
+are CSV, JSON, YAML and HTML (first table found).
+Support for Excel files is planned.
+
+=item * indent => $str
 
 Render the table with whitespace indention. Defaults to
 undefined which produces no trailing whitespace to tags.
-Useful values are some number of spaces or tabs.
-See L<HTML::Element::as_HTML()>
+Useful values are some number of spaces or tabs.  (see
+L<HTML::Element>::as_HTML).
 
-=item * encode => undef, '', or 'chars in a string'
+=item * encode => $str
 
 HTML Encode contents of td tags. Defaults to empty string
 which performs no encoding of entities. Pass a string like
 '<>&=' to perform encoding on any characters found. If the
 value is 'undef' then all unsafe characters will be
-encoded as HTML entites.
-See L<HTML::Element::as_HTML()>
+encoded as HTML entites (see L<HTML::Element>::as_HTML).
+
+=item * cache => 0 or 1
+
+Preserve data after it has been processed. Only makes
+sense when used as class method on an object.
 
 =item * matrix => 0 or 1
 
-Render the table with only td tags, no th tags.
+Render the table with only td tags, no th tags, if true.
+
+=item * headless => 0 or 1
+
+Render the table with without headings, if true.
 
 =item * table => { key => 'value' }
 
@@ -301,11 +366,33 @@ Used to generate HTML.
 
 Used for transposing data.
 
+=item L<Clone>
+
+Useful for preventing data from being clobbered.
+
 =back
 
-=head1 AUTHOR
+=head1 REQUIRES (soon to be optional)
 
-Jeff Anderson, C<< <jeffa at cpan.org> >>
+These modules are used to load data from various
+different file formats. In the future they will
+be optionally loaded as requested by the client.
+For now they are hard coded dependencies. The good
+news is they are modules that everyone should have. ;)
+
+=over 4
+
+=item L<Text::CSV>
+
+=item L<Text::CSV_XS>
+
+=item L<HTML::TableExtract>
+
+=item L<JSON>
+
+=item L<YAML>
+
+=back
 
 =head1 BUGS
 
@@ -326,14 +413,16 @@ Please report any bugs or feature requests to either
 I will be notified, and then you'll automatically be notified of progress
 on your bug as I make changes.
 
+=head1 GITHUB
+
+The Github project is L<https://github.com/jeffa/Spreadsheet-HTML>
+
 =head1 SUPPORT
 
 You can find documentation for this module with the perldoc command.
 
     perldoc Spreadsheet::HTML
 
-The Github project is L<https://github.com/jeffa/Spreadsheet-HTML>
-
 You can also look for information at:
 
 =over 4
@@ -368,6 +457,10 @@ Helped with Makefile.PL suggestions and corrections.
 
 =back
 
+=head1 AUTHOR
+
+Jeff Anderson, C<< <jeffa at cpan.org> >>
+
 =head1 LICENSE AND COPYRIGHT
 
 Copyright 2015 Jeff Anderson.
